@@ -5,11 +5,15 @@
 # Redistribution and use of this source code is licensed under
 # the BSD license. See COPYING file for license description.
 
+import base64
 import httplib
-import urllib
 import struct
 import time
 import kt_error
+try:
+    from percentcoding import quote, unquote
+except ImportError:
+    from urllib import quote, unquote
 
 try:
     import cPickle as pickle
@@ -27,7 +31,278 @@ KT_PACKER_PICKLE = 1
 KT_PACKER_JSON   = 2
 KT_PACKER_STRING = 3
 
-class ProtocolHandler:
+def _dict_to_tsv(dict):
+    return '\n'.join(quote(k) + '\t' + quote(str(v)) for (k, v) in dict.items())
+
+def _tsv_to_dict(tsv_str, content_type=''):
+    rv = {}
+    # Select the appropriate decoding function to use based on the response headers
+    if content_type.endswith('colenc=B'):
+        decode = base64.decodestring
+    elif content_type.endswith('colenc=U'):
+        decode = unquote
+    else:
+        decode = lambda x: x
+
+    for row in tsv_str.split('\n'):
+        kv = row.split('\t')
+        if len(kv) == 2:
+            rv[decode(kv[0])] = decode(kv[1])
+    return rv
+
+
+class Cursor(object):
+    cursor_id_counter = 1
+
+    def __init__(self, protocol_handler):
+        self.protocol_handler = protocol_handler
+        self.cursor_id = Cursor.cursor_id_counter
+        Cursor.cursor_id_counter += 1
+
+        self.err = kt_error.KyotoTycoonError()
+        self.pack = self.protocol_handler._pickle_packer
+        self.unpack = self.protocol_handler._pickle_unpacker
+        self.pack_type = KT_PACKER_PICKLE
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, tb):
+        # Cleanup the cursor when leaving "with" blocks
+        self.delete()
+
+    def jump(self, key=None, db=None):
+        path = '/rpc/cur_jump'
+        if db:
+            db = quote(db)
+            path += '?DB=' + db
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+        if key:
+            request_dict['key'] = key
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return True
+
+    def jump_back(self, key=None, db=None):
+        path = '/rpc/cur_jump_back'
+        if db:
+            db = quote(db)
+            path += '?DB=' + db
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+        if key:
+            request_dict['key'] = key
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return True
+
+    def step(self):
+        path = '/rpc/cur_step'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return True
+
+    def step_back(self):
+        path = '/rpc/cur_step_back'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return True
+
+    def set_value(self, value, step=False, xt=None):
+        path = '/rpc/cur_set_value'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+        request_dict['value'] = self.pack(value)
+        if step:
+            request_dict['step'] = True
+        if xt:
+            request_dict['xt'] = xt
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return True
+
+    def remove(self):
+        path = '/rpc/cur_remove'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return True
+
+    def get_key(self, step=False):
+        path = '/rpc/cur_get_key'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+        if step:
+            request_dict['step'] = True
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return _tsv_to_dict(body, res.getheader('Content-Type', ''))['key']
+
+    def get_value(self, step=False):
+        path = '/rpc/cur_get_value'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+        if step:
+            request_dict['step'] = True
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return self.unpack(_tsv_to_dict(body, res.getheader('Content-Type', ''))['value'])
+
+    def get(self, step=False):
+        path = '/rpc/cur_get'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+        if step:
+            request_dict['step'] = True
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status == 404:
+            self.err.set_error(self.err.NOTFOUND)
+            return None
+
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        res_dict = _tsv_to_dict(body, res.getheader('Content-Type', ''))
+        key = res_dict['key']
+        value = self.unpack(res_dict['value'])
+        return key, value
+
+    def seize(self):
+        path = '/rpc/cur_seize'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        res_dict = _tsv_to_dict(body, res.getheader('Content-Type', ''))
+        res_dict['key'] = res_dict['key']
+        res_dict['value'] = self.unpack(res_dict['value'])
+        return res_dict
+
+    def delete(self):
+        path = '/rpc/cur_delete'
+
+        request_dict = {}
+        request_dict['CUR'] = self.cursor_id
+
+        request_body = _dict_to_tsv(request_dict)
+        self.protocol_handler.conn.request('POST', path, body=request_body,
+                          headers=KT_HTTP_HEADER)
+
+        res, body = self.protocol_handler.getresponse()
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
+
+        self.err.set_success()
+        return True
+
+
+
+class ProtocolHandler(object):
     def __init__(self, pickle_protocol=2):
         self.err = kt_error.KyotoTycoonError()
         self.pickle_protocol = pickle_protocol
@@ -38,9 +313,18 @@ class ProtocolHandler:
     def error(self):
         return self.err
 
+    def cursor(self):
+        return Cursor(self)
+
     def open(self, host, port, timeout):
+        # Save connection parameters so the connection can be re-established
+        # on "Connection: close" response.
+        self.host = host
+        self.port = port
+        self.timeout = timeout
+
         try:
-            self.conn = httplib.HTTPConnection(host, port, timeout)
+            self.conn = httplib.HTTPConnection(host, port, timeout=timeout)
         except Exception, e:
             raise e
         return True
@@ -52,13 +336,22 @@ class ProtocolHandler:
             raise e
         return True
 
-    def echo(self):
-        self.conn.request('POST', '/rpc/echo')
+    def getresponse(self):
         res = self.conn.getresponse()
         body = res.read()
 
+        if res.will_close:
+            self.conn.close()
+            self.open(self.host, self.port, self.timeout)
+
+        return res, body
+
+    def echo(self):
+        self.conn.request('POST', '/rpc/echo')
+
+        res, body = self.getresponse()
         if res.status != 200:
-           self.err.set_error(err.EMISC)
+           self.err.set_error(self.err.EMISC)
            return False
 
         self.err.set_success()
@@ -71,15 +364,17 @@ class ProtocolHandler:
         path = key
         if db:
             path = '/%s/%s' % (db, key)
-        path = urllib.quote(path.encode('UTF-8'), safe='')
+        path = quote(path.encode('UTF-8'))
 
         self.conn.request('GET', path)
-        rv = self.conn.getresponse()
-        body = rv.read()
+        res, body = self.getresponse()
 
-        if rv.status == 404:
+        if res.status == 404:
             self.err.set_error(self.err.NOTFOUND)
             return None
+        if res.status != 200:
+            self.err.set_error(self.err.EMISC)
+            return False
 
         self.err.set_success()
         return self.unpack(body)
@@ -94,7 +389,7 @@ class ProtocolHandler:
 
         path = '/rpc/set_bulk'
         if db:
-            db = urllib.quote(db, safe='')
+            db = quote(db)
             path += '?DB=' + db
 
         request_body = ''
@@ -103,101 +398,92 @@ class ProtocolHandler:
             request_body = 'atomic\t\n'
 
         for k, v in kv_dict.items():
-            k = urllib.quote(k, safe='')
-            v = urllib.quote(self.pack(v), safe='')
+            k = quote(k)
+            v = quote(self.pack(v))
             request_body += '_' + k + '\t' + v + '\n'
 
         self.conn.request('POST', path, body=request_body,
                           headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return False
 
         self.err.set_success()
-        return int(self._tsv_to_dict(body)['num'])
+        return int(_tsv_to_dict(body, res.getheader('Content-Type', ''))['num'])
 
     def remove_bulk(self, keys, atomic, db):
-        if not isinstance(keys, list):
+        if not hasattr(keys, '__iter__'):
             self.err.set_error(self.err.LOGIC)
             return 0
 
-        if len(keys) < 1:
+        request_header = ''
+        if atomic:
+            request_header = 'atomic\t\n'
+
+        request_body = ''
+        for key in keys:
+            request_body += '_' + quote(key) + '\t\n'
+        if len(request_body) < 1:
             self.err.set_error(self.err.LOGIC)
             return 0
 
         path = '/rpc/remove_bulk'
         if db:
-            db = urllib.quote(db, safe='')
+            db = quote(db)
             path += '?DB=' + db
-
-        request_body = ''
-
-        if atomic:
-            request_body = 'atomic\t\n'
-
-        for key in keys:
-            request_body += '_' + urllib.quote(key, safe='') + '\t\n'
-
-        self.conn.request('POST', path, body=request_body,
+        self.conn.request('POST', path, body=request_header + request_body,
                           headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return False
 
         self.err.set_success()
-        return int(self._tsv_to_dict(body)['num'])
+        return int(_tsv_to_dict(body, res.getheader('Content-Type', ''))['num'])
 
     def get_bulk(self, keys, atomic, db):
-        if not isinstance(keys, list):
+        if not hasattr(keys, '__iter__'):
             self.err.set_error(self.err.LOGIC)
             return None
 
-        if len(keys) < 1:
+        request_header = ''
+        if atomic:
+            request_header = 'atomic\t\n'
+
+        request_body = ''
+        for key in keys:
+            request_body += '_' + quote(key) + '\t\n'
+
+        if len(request_body) < 1:
             self.err.set_error(self.err.LOGIC)
             return {}
 
         path = '/rpc/get_bulk'
         if db:
-            db = urllib.quote(db, safe='')
+            db = quote(db)
             path += '?DB=' + db
-
-        request_body = ''
-
-        if atomic:
-            request_body = 'atomic\t\n'
-
-        for key in keys:
-            request_body += '_' + urllib.quote(key, safe='') + '\t\n'
-
-        self.conn.request('POST', path, body=request_body,
+        self.conn.request('POST', path, body=request_header + request_body,
                           headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return None
 
         rv = {}
-        res_dict = self._tsv_to_dict(body)
+        res_dict = _tsv_to_dict(body, res.getheader('Content-Type', ''))
         n = res_dict.pop('num')
 
-        if n == 0:
+        if n == '0':
             self.err.set_error(self.err.NOTFOUND)
-            return None
+            return {}
 
         for k, v in res_dict.items():
             if v is not None:
-                rv[urllib.unquote(k[1:])] = self.unpack(urllib.unquote(v))
+                rv[k[1:]] = self.unpack(v)
 
         self.err.set_success()
         return rv
@@ -210,30 +496,28 @@ class ProtocolHandler:
         path = key
         if db:
             path = '/%s/%s' % (db, key)
-        path = urllib.quote(path.encode('UTF-8'), safe='')
+        path = quote(path.encode('UTF-8'))
 
         self.conn.request('GET', path)
-        rv = self.conn.getresponse()
-        buf = rv.read()
 
-        if rv.status != 200:
+        res, body = self.getresponse()
+        if res.status != 200:
             self.err.set_error(self.err.NOTFOUND)
             return None
 
         self.err.set_success()
-        return struct.unpack('>q', buf)[0]
+        return struct.unpack('>q', body)[0]
 
     def vacuum(self, db):
         path = '/rpc/vacuum'
 
         if db:
-            db = urllib.quote(db, safe='')
+            db = quote(db)
             path += '?DB=' + db
 
         self.conn.request('GET', path)
-        res = self.conn.getresponse()
-        body = res.read()
 
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
 
@@ -254,23 +538,21 @@ class ProtocolHandler:
         if db:
             request_dict['DB'] = db
 
-        request_body = self._dict_to_tsv(request_dict)
+        request_body = _dict_to_tsv(request_dict)
         self.conn.request('POST', '/rpc/match_prefix',
                           body=request_body, headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return False
 
-        res_dict = self._tsv_to_dict(body)
+        res_dict = _tsv_to_dict(body, res.getheader('Content-Type', ''))
         n = res_dict.pop('num')
 
-        if n == 0:
+        if n == '0':
             self.err.set_error(self.err.NOTFOUND)
-            return None
+            return []
 
         for k in res_dict.keys():
             rv.append(k[1:])
@@ -291,21 +573,19 @@ class ProtocolHandler:
         if max:
             request_dict['max'] = max
 
-        request_body = self._dict_to_tsv(request_dict)
+        request_body = _dict_to_tsv(request_dict)
         self.conn.request('POST', path, body=request_body,
                           headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return None
 
         rv = []
-        res_dict = self._tsv_to_dict(body)
+        res_dict = _tsv_to_dict(body, res.getheader('Content-Type', ''))
 
-        if res_dict.pop('num') < 1:
+        if res_dict.pop('num') == '0':
             self.err.set_error(self.err.NOTFOUND)
             return []
 
@@ -322,13 +602,12 @@ class ProtocolHandler:
 
         if db:
             key = '/%s/%s' % (db, key)
-        key = urllib.quote(key.encode('UTF-8'), safe='')
+        key = quote(key.encode('UTF-8'))
         value = self.pack(value)
 
         self.err.set_success()
 
         status = self._rest_put('set', key, value, expire)
-
         if status != 201:
             self.err.set_error(self.err.EMISC)
             return False
@@ -344,7 +623,7 @@ class ProtocolHandler:
         if db:
             key = '/%s/%s' % (db, key)
 
-        key = urllib.quote(key.encode('UTF-8'), safe='')
+        key = quote(key.encode('UTF-8'))
         value = self.pack(value)
         status = self._rest_put('add', key, value, expire)
 
@@ -367,20 +646,18 @@ class ProtocolHandler:
         request_dict = { 'key': key }
 
         if old_val:
-            request_dict['oval'] = urllib.quote(self.pack(old_val), safe='')
+            request_dict['oval'] = self.pack(old_val)
         if new_val:
-            request_dict['nval'] = urllib.quote(self.pack(new_val), safe='')
+            request_dict['nval'] = self.pack(new_val)
         if expire:
             request_dict['xt'] = expire
 
-        request_body = self._dict_to_tsv(request_dict)
+        request_body = _dict_to_tsv(request_dict)
 
         self.conn.request('POST', path, body=request_body,
                           headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return False
@@ -396,12 +673,11 @@ class ProtocolHandler:
         if db:
             key = '/%s/%s' % (db, key)
 
-        key = urllib.quote(key.encode('UTF-8'), safe='')
+        key = quote(key.encode('UTF-8'))
         self.conn.request('DELETE', key)
-        rv = self.conn.getresponse()
-        body = rv.read()
 
-        if rv.status != 204:
+        res, body = self.getresponse()
+        if res.status != 204:
             self.err.set_error(self.err.NOTFOUND)
             return False
 
@@ -416,7 +692,7 @@ class ProtocolHandler:
         if db:
             key = '/%s/%s' % (db, key)
 
-        key = urllib.quote(key.encode('UTF-8'), safe='')
+        key = quote(key.encode('UTF-8'))
         value = self.pack(value)
         status = self._rest_put('replace', key, value, expire)
 
@@ -463,15 +739,13 @@ class ProtocolHandler:
         self.conn.request('POST', path, body=request_body,
                           headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return None
 
         self.err.set_success()
-        return int(self._tsv_to_dict(body)['num'])
+        return int(_tsv_to_dict(body, res.getheader('Content-Type', ''))['num'])
 
     def increment_double(self, key, delta, expire, db):
         if key is None:
@@ -487,57 +761,49 @@ class ProtocolHandler:
         self.conn.request('POST', path, body=request_body,
                           headers=KT_HTTP_HEADER)
 
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return None
 
         self.err.set_success()
-        return float(self._tsv_to_dict(body)['num'])
+        return float(_tsv_to_dict(body, res.getheader('Content-Type', ''))['num'])
 
     def report(self):
         self.conn.request('GET', '/rpc/report')
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return None
 
         self.err.set_success()
-        return self._tsv_to_dict(body)
+        return _tsv_to_dict(body, res.getheader('Content-Type', ''))
 
     def status(self, db=None):
         url = '/rpc/status'
 
         if db:
-            db = urllib.quote(db, safe='')
+            db = quote(db)
             url += '?DB=' + db
 
         self.conn.request('GET', url)
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return None
 
         self.err.set_success()
-        return self._tsv_to_dict(body)
+        return _tsv_to_dict(body, res.getheader('Content-Type', ''))
 
     def clear(self, db=None):
         url = '/rpc/clear'
 
         if db:
-            db = urllib.quote(db, safe='')
+            db = quote(db)
             url += '?DB=' + db
 
         self.conn.request('GET', url)
-        res = self.conn.getresponse()
-        body = res.read()
-
+        res, body = self.getresponse()
         if res.status != 200:
             self.err.set_error(self.err.EMISC)
             return False
@@ -557,17 +823,6 @@ class ProtocolHandler:
             return None
         return int(st['size'])
 
-    def _dict_to_tsv(self, dict):
-        return '\n'.join(k + '\t' + str(v) for (k, v) in dict.items())
-
-    def _tsv_to_dict(self, tsv_str):
-        rv = {}
-        for row in tsv_str.split('\n'):
-            kv = row.split('\t')
-            if len(kv) == 2:
-                rv[kv[0]] = kv[1]
-        return rv
-
     def _rest_put(self, operation, key, value, expire):
         headers = { 'X-Kt-Mode' : operation }
         if expire != None:
@@ -575,9 +830,8 @@ class ProtocolHandler:
             headers["X-Kt-Xt"] = str(expire)
 
         self.conn.request('PUT', key, value, headers)
-        rv = self.conn.getresponse()
-        body = rv.read()
-        return rv.status
+        res, body = self.getresponse()
+        return res.status
 
     def _pickle_packer(self, data):
         return pickle.dumps(data, self.pickle_protocol)
